/***************************************************************************
 *   Copyright (C) 2007 by Marco Lorrai                                    *
 *   marco.lorrai@abbeynet.it                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "frame.h"
#include "setting.h"
#include "pwc-ioctl.h"
#include "wxcam.h"
#include "progressdlg.h"
#include "filters.h"
#include <wx/wfstream.h>
#include <wx/thread.h>
#include <wx/progdlg.h>
#include <sys/time.h>
#include <time.h>

BEGIN_EVENT_TABLE( Frame, wxWindow )
EVT_UPDATE_UI( PROGRESS_DIALOG_START, Frame::OnUpdateProgress )
EVT_MENU( PROGRESS_DIALOG_START, Frame::OnStartProgress )
EVT_MENU( PROGRESS_DIALOG_EVENT, Frame::OnProgressEvent )
EVT_MENU( STOP_REC, Frame::OnRecThreadTerminated )
EVT_MENU( AUDIO_ERR, Frame::OnAudioError )
EVT_PAINT( Frame::OnPaint )
EVT_ERASE_BACKGROUND( Frame::OnEraseBackground )
//EVT_IDLE(Frame::OnIdle)
END_EVENT_TABLE()

using namespace std;

Device cam;

Frame::Frame( const string &resolution, wxCam* main, wxWindow* parent, wxWindowID id,
        const wxPoint& pos, const wxSize& size, long style )
        : wxWindow( parent, id, pos, size, style ) 
 {
    mainFrame = main;
    int w, h;
    GetResolution( resolution, w, h );
    width=w;
    height=h;
    snapshot = false;
    started = false;
    fullScreen = false;
    rstatus = REC_NONE;
    rec = NULL;
    dialog = ( ProgressDlg* ) NULL;
    wxInitAllImageHandlers();
    OpenDevice();    
    busy=false;
}

Frame::~Frame()
{}

void Frame::OpenDevice() 
{
    int frameFormat;
    if ( Setting::GetInstance()->GetFrameFormat() == "YUV420P" )
        frameFormat = VIDEO_PALETTE_YUV420P;
    else if( Setting::GetInstance()->GetFrameFormat() == "YUV420" )
        frameFormat = VIDEO_PALETTE_YUV420;
    else if( Setting::GetInstance()->GetFrameFormat() == "YUYV" )
        frameFormat = VIDEO_PALETTE_YUYV;
    else if( Setting::GetInstance()->GetFrameFormat() == "RGB24" )
        frameFormat = VIDEO_PALETTE_RGB24;
    else if( Setting::GetInstance()->GetFrameFormat() == "RGB32" )
        frameFormat = VIDEO_PALETTE_RGB32;
    else
        frameFormat = 0 /*AUTO*/;
    if ( !cam.openDevice( width, height, frameFormat ) ) {
        deviceOpen = false;
        wxString msg = _( "Cannot open " );
        msg += wxString(Setting::GetInstance()->GetDeviceFile().c_str(), wxConvUTF8);
        msg += wxString( ".\n", wxConvUTF8 ) + _( "Please check if your system has the correct driver for your webcam, or change the webcam device in settings->preferences." );
        wxMessageBox( msg, _( "Error opening device" ), wxICON_ERROR );
    }
    else
        deviceOpen = true;
    buffer=NULL;
}

void Frame::Started()  //now we can acquire frame!
{
    Connect( wxEVT_IDLE, wxIdleEventHandler( Frame::OnIdle ) );
}

bool Frame::isPWC() 
{
    return cam.isPWC();
}

void Frame::TakeSnapshot() 
{
    snapshot = true;
}

void Frame::startRecording() 
{
    if(rstatus == REC_NONE)
        rstatus = REC_STARTING;
}

void Frame::stopRecording() 
{
    if(rstatus == REC_CONTINUING)
        rstatus = REC_STOPPING;
}

void Frame::setFullscreen( bool enable ) 
{
    fullScreen=enable;
}

void Frame::OnPaint( wxPaintEvent& event ) 
{
    int w, h, frameW, frameH;
    wxPaintDC dc( this );
    dc.GetSize( &w, &h );
    frameW = w;
    frameH = h;
    dc.SetBrush( wxBrush( *wxWHITE ) );
    dc.SetPen( wxPen( *wxWHITE, 1, wxSOLID ) );
    dc.DrawRectangle( 0, 0, w, h );
    if ( buffer ) {
        int x, y;
        x = w/2 - ( width + 2 ) /2;
        y = h/2 - ( height + 2 ) /2;
        w = width;
        h = height;
        wxImage wximg( w, h );

        if ( rstatus == REC_STARTING ) {
            mainFrame->enableRec( false );
            compression = Setting::GetInstance()->GetCompression();
            if(compression == "none") 
                rec = new AviUncompressed();
            else {
                rec = new Xvid();
                float level = (float)Setting::GetInstance()->GetCompressionLevel();
                static_cast<Xvid*>(rec)->setCompressionLevel(level/(float)100);
            }
            string filename = Setting::GetInstance()->GetVideoFile(false);
            unsigned int gw, gh, freq;
            cam.getResolution( gw, gh, freq );            
            if ( !cam.isPWC() )
                freq = assessFrequency;            
            rec->setParameters( this, filename, freq, gw, gh );
            if(compression == "none") {
                if ( static_cast<AviUncompressed*>(rec)->Create() != wxTHREAD_NO_ERROR ) {
                    wxLogError( _( "Cannot create video recording thread!" ) );
                    return;
                }

                static_cast<AviUncompressed*>(rec)->Run();
            }

            rec->addFrame( buffer );
            rstatus = REC_CONTINUING;
        }
        else if ( rstatus == REC_CONTINUING ) {
            if ( rec )
                rec->addFrame( buffer );
        }
        else if ( rstatus == REC_STOPPING ) {
            if ( rec ) {
                setBusy( true );
                rec->record();
            }
            rstatus = REC_NONE;
        }

        wximg.SetData( ( unsigned char* ) buffer );

        if ( snapshot ) {
            snapshot = false;
            wxFileOutputStream stream( wxString(Setting::GetInstance()->GetSnapshotFile( false ).c_str(), wxConvUTF8) );
            wximg.SaveFile( stream, wxBITMAP_TYPE_TIF );
        }
        if ( fullScreen ) {
            wximg.Rescale( frameH*4/3, frameH );
            int diff = frameW - frameH*4/3;
            wxBitmap bit( wximg );
            dc.DrawBitmap( bit, diff/2 , 0, true );
        }
        else {
            wxBitmap bit( wximg );
            dc.SetPen( wxPen( *wxBLACK, 1, wxSOLID ) );
            dc.DrawRectangle( x, y, w + 2, h + 2 );
            dc.DrawBitmap( bit, 1 + x , 1 + y, true );
        }
    }
}

void Frame::OnEraseBackground( wxEraseEvent &event )
{}

bool Frame::isOpen() 
{
    return deviceOpen;
}

bool Frame::SetResolution( const std::string &resolution, const std::string &frameRate ) 
{
    int w, h;
    GetResolution( resolution, w, h );
    int freq=0;
    if( cam.isPWC() )
        freq=atoi(frameRate.c_str());
    if ( !cam.setResolution( w, h, freq ) && deviceOpen ) {
        wxMessageBox( _( "Frame rate/resolution setting not supported.\nPlease try a lower frame rate o resolution." ), _( "Error applying settings" ), wxICON_WARNING );
        return false;
    }
    else {
        Setting::GetInstance()->SetFrameRate( frameRate.c_str() );
        Setting::GetInstance()->SetResolution( resolution.c_str() );
        width=w;
        height=h;
        return true;
    }
}

void Frame::OnIdle( wxIdleEvent& event ) 
{
    static struct timeval tv;
    static bool show = false;
    if ( deviceOpen && cam.getFrame( &buffer ) ) {
        applyFilters();
        struct timeval tv_old = tv;
        gettimeofday( &tv, NULL );
        int delta = ( tv.tv_sec - tv_old.tv_sec ) * 1000000 + ( tv.tv_usec - tv_old.tv_usec );
        double sec = ( double ) delta / ( double ) 1000000;
        float freq = ( float ) 1/ ( float ) sec;
        updateFrequency( freq );
        Refresh();
        Update();
        buffer=NULL;
    }
    else {
        if ( deviceOpen && !show ) {
            show=true;
            Disconnect( wxEVT_IDLE, wxIdleEventHandler( Frame::OnIdle ) );
            wxMessageBox( _( "An error has occured during frame capture.\nPlease check the \"frame format\" options in the preferences menu." ), _( "Error capturing frame" ), wxICON_ERROR );
            return;
        }
    }
    if ( !isBusy() )
        event.RequestMore();
    event.Skip();
}

void Frame::GetResolution( const std::string &resolution, int &w, int &h ) 
{
    wxString tmp = wxString(resolution.c_str(), wxConvUTF8), res = wxString(resolution.c_str(), wxConvUTF8);
    int pos = res.Find( 'x' );
    h = atoi( (const char*)res.Remove( 0, pos + 1 ).mb_str(wxConvUTF8) );
    w = atoi( (const char*)tmp.Remove( pos ).mb_str(wxConvUTF8) );
}

void Frame::updateFrequency( float freq ) 
{
    freqList.push_back( freq );
    if ( freqList.size() == 11 )
        freqList.pop_front();
    int c=0;
    float mean=0;
    for ( std::list<float>::iterator it=freqList.begin(); it!=freqList.end(); ++it ) {
        mean += *it;
        c++;
    }
    mean /= ( float ) c;
    assessFrequency = ( int ) ( mean + 0.5 );
}

void Frame::OnProgressEvent( wxCommandEvent& event ) 
{
    int n = event.GetInt();
    cout<<"OnProgressEvent: n="<<n<<endl;
    if (n == -1) {
        dialog->Hide();
        dialog->Destroy();
        dialog = (ProgressDlg *)NULL;
        // the dialog is aborted because the event came from another thread, so
        // we may need to wake up the main event loop for the dialog to be
        // really closed

        wxWakeUpIdle();
    }
    else
        dialog->Update(n);
}

void Frame::OnUpdateProgress( wxUpdateUIEvent& event ) 
{
    event.Enable( isBusy() );
}

void Frame::OnStartProgress( wxCommandEvent& event ) 
{
    cout<<"OnStartProgress: Dialog started"<<endl;
//    dialog = new wxProgressDialog
//                    (
//                     _T(_("Recording file")),
//                     _T(_("Please wait until the file is saved...")),
//                     100,
//                     this,
//                     wxPD_AUTO_HIDE |
//                     wxPD_APP_MODAL |
//                     wxPD_ELAPSED_TIME |
//                     wxPD_ESTIMATED_TIME |
//                     wxPD_REMAINING_TIME
//                    );
    dialog = new ProgressDlg(this);
    dialog->Show();
    cout<<"OnStartProgress: Dialog created"<<endl;
}

void Frame::OnRecThreadTerminated( wxCommandEvent& event ) 
{
    int n = event.GetInt();
    if ( n == 1 ) {
        wxString message =  _( "An error has occured opening file " );
        message += wxString(Setting::GetInstance()->GetVideoFile().c_str(), wxConvUTF8);
        wxMessageBox( message, _( "Error opening file" ), wxICON_WARNING );
        if(compression == "none")
            static_cast<AviUncompressed*>(rec)->Delete();
        rec = NULL;
        mainFrame->closeRecWindow();
    }
    if ( n == 2 ) {
        wxString message =  _( "You system is slow" );
        wxMessageBox( message, _( "Queue full" ), wxICON_WARNING );
        mainFrame->closeRecWindow();
        stopRecording();
    }    
    
    mainFrame->enableRec( true );
    setBusy( false );
    wxWakeUpIdle();
}

void Frame::OnAudioError( wxCommandEvent& event )
{
    int n = event.GetInt();
    if ( n == 1 ) {
        wxString message = _( "Cannot open /dev/dsp.\n" );
        message += _( "Video file will be recorded without audio track.\n" );
        wxMessageBox( message, _( "Error recording audio" ), wxICON_ERROR );       
    }
}

bool Frame::isBusy() 
{
    return busy;
}

void Frame::setBusy( bool b ) 
{
    busy=b;
    if ( !busy )
        Connect( wxEVT_IDLE, wxIdleEventHandler( Frame::OnIdle ) );
    else
        Disconnect( wxEVT_IDLE, wxIdleEventHandler( Frame::OnIdle ) );
}

void Frame::applyFilters( )
{
    int size = width * height *3; 
    if(Setting::GetInstance()->GetAdjustColors())
        cam.setAdjustColors(true);
    else
        cam.setAdjustColors(false);
    if(Setting::GetInstance()->GetNegative())
        negative( (unsigned char*)buffer, size);
    if(Setting::GetInstance()->GetMonochrome())
        monochrome((unsigned char*)buffer, size);
    if(Setting::GetInstance()->GetUpturned())
        upturned( (unsigned char*)buffer, size);
    if(Setting::GetInstance()->GetMirror())
        mirror( (unsigned char*)buffer, size, width);
    if(Setting::GetInstance()->GetEdge())
        edge((unsigned char*)buffer, size, width);
}
