/***************************************************************************
 *   Copyright (C) 2007 by Marco Lorrai                                    *
 *   marco.lorrai@abbeynet.it                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "setting.h"
#include "v4l2.h"
#include "ccvt.h"
#include "filters.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>     
#include <assert.h>
#include <fcntl.h>              /* low-level i/o */
#include <unistd.h> 
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h>          /* for videodev2.h */

#include <linux/videodev2.h>


VideoForLinux2::VideoForLinux2()
{
    dev_name = Setting::GetInstance()->GetDeviceFile();
    buffers = NULL;
    n_buffers = 0;
    first = true;
    pixelFormat = 0;
    isMapped = false;
    setControls();
}
   
VideoForLinux2::~VideoForLinux2()
{
    unsigned int i;
    enum v4l2_buf_type type;
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (ioctl (fd, VIDIOC_STREAMOFF, &type) == -1)
        perror("VIDIOC_STREAMOFF");
    
     for (i = 0; i < n_buffers; ++i)
         if (munmap (buffers[i].start, buffers[i].length) == -1)
             perror("munmap");
    
    free(buffers);
}

bool VideoForLinux2::mMap(int w, int h) {
    struct v4l2_cropcap cropcap;
    struct v4l2_crop crop;
    struct v4l2_format fmt;
    unsigned int min;
    
    width = w;
    height = h;
        
    memset(&cropcap, 0, sizeof(struct v4l2_cropcap));
    cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    
    if (!ioctl(fd, VIDIOC_CROPCAP, &cropcap)) {
        crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        crop.c = cropcap.defrect; /* reset to default */
        
        if (ioctl(fd, VIDIOC_S_CROP, &crop) == -1) {
            switch (errno) {
                case EINVAL:
                    /* Cropping not supported. */
                    break;
                default:
                    /* Errors ignored. */
                    break;
            }
        }
    }
    else {
                /* Errors ignored. */
    }
    
    memset(&fmt, 0, sizeof(struct v4l2_format));
    fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width       = w;
    fmt.fmt.pix.height      = h;
    
    if(frameFormat==VIDEO_PALETTE_RGB24) {
        fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24; 
        pixelFormat = V4L2_PIX_FMT_BGR24;
        std::cout<<"Using V4L2_PIX_FMT_BGR24 pixel format"<<std::endl;
    }
    else if(frameFormat==VIDEO_PALETTE_YUYV) {
        fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
        pixelFormat = V4L2_PIX_FMT_YUYV;
        std::cout<<"Using V4L2_PIX_FMT_YUYV pixel format"<<std::endl;
    }
    else if(frameFormat==VIDEO_PALETTE_YUV420P) {
        fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420;
        pixelFormat = V4L2_PIX_FMT_YUV420;
        std::cout<<"Using V4L2_PIX_FMT_YUV420 pixel format"<<std::endl;
    }
    else { //AUTO or YUV420
        struct v4l2_fmtdesc argp;
        memset(&argp, 0, sizeof(struct v4l2_fmtdesc));
        argp.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        std::cout<<"Determining pixel format..."<<std::endl;
        for(int index=0; index < 4; index++) {            
            argp.index=index;
            if (ioctl(fd, VIDIOC_ENUM_FMT, &argp) == -1) {
                //perror("VIDIOC_ENUM_FMT");
                break;
            }
            else {
                std::cout<<"pixel format: "<<argp.description<<std::endl;
                if(argp.pixelformat == V4L2_PIX_FMT_YUV420) {
                    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420;
                    pixelFormat = V4L2_PIX_FMT_YUV420;
                    std::cout<<"Found V4L2_PIX_FMT_YUV420 pixel format"<<std::endl;
                }
                else if(argp.pixelformat == V4L2_PIX_FMT_YUYV) {
                    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
                    pixelFormat = V4L2_PIX_FMT_YUYV;
                    std::cout<<"Found V4L2_PIX_FMT_YUYV pixel format"<<std::endl;
                }                
                else if(argp.pixelformat == V4L2_PIX_FMT_BGR24) {
                    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
                    pixelFormat = V4L2_PIX_FMT_BGR24;
                    std::cout<<"Found V4L2_PIX_FMT_BGR24 pixel format"<<std::endl;
                }
            }                
        }
    }
    
    if(!pixelFormat) {
        std::cout<<"Could not determine pixel format"<<std::endl;
        return false;
    }
    
    fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
    
    if (ioctl (fd, VIDIOC_S_FMT, &fmt) == -1) {
        perror("VIDIOC_S_FMT");
        return false;
    }
    
    /* Buggy driver paranoia. */
    min = fmt.fmt.pix.width * 2;
    if (fmt.fmt.pix.bytesperline < min)
        fmt.fmt.pix.bytesperline = min;
    min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
    if (fmt.fmt.pix.sizeimage < min)
        fmt.fmt.pix.sizeimage = min;
    
    return init_mmap();
    
}

bool VideoForLinux2::init_mmap()
{
    struct v4l2_requestbuffers req;

    memset(&req, 0, sizeof(struct v4l2_requestbuffers));
    
    req.count               = 4;
    req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory              = V4L2_MEMORY_MMAP;
    
    if (ioctl (fd, VIDIOC_REQBUFS, &req) == -1) {
        if (EINVAL == errno) {
            fprintf (stderr, "%s does not support "
                    "memory mapping\n", dev_name.c_str());
            return false;
        } 
        else {
            perror("VIDIOC_REQBUFS");
            return false;
        }
    }
    
    if (req.count < 2) {
        fprintf(stderr, "Insufficient buffer memory on %s\n",
                dev_name.c_str());
        return false;
    }
    
    if(buffers) {
        for (unsigned int i = 0; i < n_buffers; ++i)
            if (munmap (buffers[i].start, buffers[i].length) == -1)
                perror("munmap");
        free(buffers);
    }
    buffers = (buffer*)calloc (req.count, sizeof (*buffers));
    if (!buffers) { 
        fprintf (stderr, "Out of memory\n");
        return false;
    }
    
    for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
        struct v4l2_buffer buf;

        memset(&buf, 0, sizeof(struct v4l2_buffer));

        buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory      = V4L2_MEMORY_MMAP;
        buf.index       = n_buffers;

        if (ioctl (fd, VIDIOC_QUERYBUF, &buf) == -1) {
            perror("VIDIOC_QUERYBUF");
            return false;
        }

        buffers[n_buffers].length = buf.length;
        buffers[n_buffers].start =
                mmap (NULL /* start anywhere */,
                      buf.length,
                      PROT_READ | PROT_WRITE /* required */,
                      MAP_SHARED /* recommended */,
                      fd, buf.m.offset);

        if (MAP_FAILED == buffers[n_buffers].start) {
            perror("mmap");
            return false;
        }
    }
    isMapped = true;
    return true;
}

bool VideoForLinux2::getFrame(char **buffer)
{
    unsigned int i;
    enum v4l2_buf_type type;
    struct v4l2_buffer buf;
    
    if(!isMapped) {
        std::cout<<"Device not mapped"<<std::endl;
        return false;
    }
    
    if(first) {
        for (i = 0; i < n_buffers; ++i) {
            memset(&buf, 0, sizeof(struct v4l2_buffer));

            buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory      = V4L2_MEMORY_MMAP;
            buf.index       = i;

            if (ioctl (fd, VIDIOC_QBUF, &buf) == -1) {
                perror("VIDIOC_QBUF");
                return false;
            }
        }

        type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (ioctl (fd, VIDIOC_STREAMON, &type) == -1) {
            perror("VIDIOC_STREAMON");
            return false;
        }
    }
    
    first = false;
    memset(&buf, 0, sizeof(struct v4l2_buffer));
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

    if (ioctl (fd, VIDIOC_DQBUF, &buf) == -1) {
        switch (errno) {
            case EAGAIN:
                return false;
                
                case EIO:
                        /* Could ignore EIO, see spec. */

                        /* fall through */

                default:
                    perror("VIDIOC_DQBUF");
                    return false;
                }
    }

    assert(buf.index < n_buffers);       

    *buffer = ( char* ) malloc( width*height*3 );

    if(pixelFormat == V4L2_PIX_FMT_YUV420) {
        if(adjustColors)
            ccvt_420p_bgr24( width, height, ( const unsigned char* ) buffers[buf.index].start, ( unsigned char* ) *buffer );
        else
            ccvt_420p_rgb24( width, height, ( const unsigned char* ) buffers[buf.index].start, ( unsigned char* ) *buffer );             
    }
    else if(pixelFormat == V4L2_PIX_FMT_YUYV) {
        if(adjustColors)
            ccvt_yuyv_bgr24( width, height, ( const unsigned char* ) buffers[buf.index].start, ( unsigned char* ) *buffer );
        else
            ccvt_yuyv_rgb24( width, height, ( const unsigned char* ) buffers[buf.index].start, ( unsigned char* ) *buffer ); 
    }
    else if(pixelFormat == V4L2_PIX_FMT_BGR24) {
        if(!adjustColors)
            convert2bgr( (unsigned char*)*buffer, width*height*3);   
    }
    else {
        std::cout<<"Unknown pixel format"<<std::endl;
        return false;
    }   

    if (ioctl (fd, VIDIOC_QBUF, &buf) == -1) {
        perror("VIDIOC_QBUF");
        return false;
    }
    
    return true;
}

bool VideoForLinux2::setResolution(unsigned int width, unsigned int height, unsigned int frameRate)
{   
    frameRate=0; /*avoiding warning*/
    return mMap(width, height);
}

bool VideoForLinux2::getResolution(unsigned int &width, unsigned int &height, unsigned int &frameRate)
{
    struct v4l2_format fmt;
    memset(&fmt, 0, sizeof(struct v4l2_format));
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    frameRate=0; /*ignored*/
    if (ioctl (fd, VIDIOC_G_FMT, &fmt) == -1) {
        perror("VIDIOC_G_FMT");
        return false;
    }
    width = fmt.fmt.pix.width;
    height = fmt.fmt.pix.height;
    printf("width: %d, height %d\n", width, height);
    return true;
}

bool VideoForLinux2::setBrightness(unsigned int value)
{
    struct v4l2_control control;
    memset (&control, 0, sizeof (control));
    control.id = V4L2_CID_BRIGHTNESS;
    control.value = value;
    
    if (ioctl (fd, VIDIOC_S_CTRL, &control) == -1) {
        perror ("VIDIOC_S_CTRL");
        return false;
    }
    
    return true;    
}

bool VideoForLinux2::getBrightness(unsigned int &value)
{
    struct v4l2_control control;
    memset (&control, 0, sizeof (control));
    control.id = V4L2_CID_BRIGHTNESS;

    if (ioctl (fd, VIDIOC_G_CTRL, &control) == 0) {
        value = control.value;
        /* Ignore if V4L2_CID_BRIGHTNESS is unsupported */
    } 
    else if (errno != EINVAL) {
        perror ("VIDIOC_G_CTRL");
        return false;
    }        
    return true;
}

bool VideoForLinux2::setContrast(unsigned int value)
{
    struct v4l2_control control;
    memset (&control, 0, sizeof (control));
    control.id = V4L2_CID_CONTRAST;
    control.value = value;
    
    if (ioctl (fd, VIDIOC_S_CTRL, &control) == -1) {
        perror ("VIDIOC_S_CTRL");
        return false;
    }
    
    return true;      
}

bool VideoForLinux2::getContrast(unsigned int &value)
{
    struct v4l2_control control;
    memset (&control, 0, sizeof (control));
    control.id = V4L2_CID_CONTRAST;

    if (ioctl (fd, VIDIOC_G_CTRL, &control) == 0) {
        value = control.value;
        /* Ignore if V4L2_CID_CONTRAST is unsupported */
    } 
    else if (errno != EINVAL) {
        perror ("VIDIOC_G_CTRL");
        return false;
    }        
    return true;
}

bool VideoForLinux2::setGamma(unsigned int value)
{
    struct v4l2_control control;
    memset (&control, 0, sizeof (control));
    control.id = V4L2_CID_GAMMA;
    control.value = value;
    
    if (ioctl (fd, VIDIOC_S_CTRL, &control) == -1) {
        perror ("VIDIOC_S_CTRL");
        return false;
    }
    
    return true;    
}

bool VideoForLinux2::getGamma(unsigned int &value)
{
    struct v4l2_control control;
    memset (&control, 0, sizeof (control));
    control.id = V4L2_CID_GAMMA;

    if (ioctl (fd, VIDIOC_G_CTRL, &control) == 0) {
        value = control.value;
        /* Ignore if V4L2_CID_GAMMA is unsupported */
    } 
    else if (errno != EINVAL) {
        perror ("VIDIOC_G_CTRL");
        return false;
    }        
    return true;
}

bool VideoForLinux2::setSaturation(unsigned int value)
{
    struct v4l2_control control;
    memset (&control, 0, sizeof (control));
    control.id = V4L2_CID_SATURATION;
    control.value = value;
    
    if (ioctl (fd, VIDIOC_S_CTRL, &control) == -1) {
        perror ("VIDIOC_S_CTRL");
        return false;
    }
    
    return true;    
}

bool VideoForLinux2::getSaturation(unsigned int &value)
{
    struct v4l2_control control;
    memset (&control, 0, sizeof (control));
    control.id = V4L2_CID_SATURATION;

    if (ioctl (fd, VIDIOC_G_CTRL, &control) == 0) {
        value = control.value;
        /* Ignore if V4L2_CID_SATURATION is unsupported */
    } 
    else if (errno != EINVAL) {
        perror ("VIDIOC_G_CTRL");
        return false;
    }        
    return true;
}

void VideoForLinux2::setControls()
{
    struct v4l2_queryctrl queryctrl;

    enumerateControls();
    
    /*BRIGHTNESS*/
    memset (&queryctrl, 0, sizeof (queryctrl));
    queryctrl.id = V4L2_CID_BRIGHTNESS;
    queryctrl.minimum = 0;
    queryctrl.maximum = 65535;

    if (ioctl (fd, VIDIOC_QUERYCTRL, &queryctrl) ==-1) {
        if (errno != EINVAL) {
                perror ("VIDIOC_QUERYCTRL");
                //return false;
        } 
        else {
            printf ("V4L2_CID_BRIGHTNESS is not supported\n");
        }
    } 
    else if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
        printf ("V4L2_CID_BRIGHTNESS is not supported\n");
    } 
    
    /*CONTRAST*/
    memset (&queryctrl, 0, sizeof (queryctrl));
    queryctrl.id = V4L2_CID_CONTRAST;
    queryctrl.minimum = 0;
    queryctrl.maximum = 65535;

    if (ioctl (fd, VIDIOC_QUERYCTRL, &queryctrl) ==-1) {
        if (errno != EINVAL) {
                perror ("VIDIOC_QUERYCTRL");
                //return false;
        } 
        else {
            printf ("V4L2_CID_CONTRAST is not supported\n");
        }
    } 
    else if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
        printf ("V4L2_CID_CONTRAST is not supported\n");
    } 
    
    /*GAMMA*/
    memset (&queryctrl, 0, sizeof (queryctrl));
    queryctrl.id = V4L2_CID_GAMMA;
    queryctrl.minimum = 0;
    queryctrl.maximum = 65535;

    if (ioctl (fd, VIDIOC_QUERYCTRL, &queryctrl) ==-1) {
        if (errno != EINVAL) {
                perror ("VIDIOC_QUERYCTRL");
                //return false;
        } 
        else {
            printf ("V4L2_CID_GAMMA is not supported\n");
        }
    } 
    else if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
        printf ("V4L2_CID_GAMMA is not supported\n");
    } 
    
    /*SATURATION*/
    memset (&queryctrl, 0, sizeof (queryctrl));
    queryctrl.id = V4L2_CID_SATURATION;
    queryctrl.minimum = 0;
    queryctrl.maximum = 65535;

    if (ioctl (fd, VIDIOC_QUERYCTRL, &queryctrl) ==-1) {
        if (errno != EINVAL) {
                perror ("VIDIOC_QUERYCTRL");
                //return false;
        } 
        else {
            printf ("V4L2_CID_SATURATION is not supported\n");
        }
    } 
    else if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
        printf ("V4L2_CID_SATURATION is not supported\n");
    } 
    
}

void VideoForLinux2::enumerateMenu(struct v4l2_queryctrl &queryctrl, struct v4l2_querymenu querymenu)
{
    printf ("  Menu items:\n");

    memset (&querymenu, 0, sizeof (querymenu));
    querymenu.id = queryctrl.id;

    for (querymenu.index = queryctrl.minimum;
         querymenu.index <= queryctrl.maximum;
          querymenu.index++) {
            if (0 == ioctl (fd, VIDIOC_QUERYMENU, &querymenu)) {
                    printf ("  %s\n", querymenu.name);
            } else {
                    perror ("VIDIOC_QUERYMENU");
                    exit (EXIT_FAILURE);
            }
    }
}

void VideoForLinux2::enumerateControls()
{
    struct v4l2_queryctrl queryctrl;
    struct v4l2_querymenu querymenu;
    
    memset (&queryctrl, 0, sizeof (queryctrl));
    
    for (queryctrl.id = V4L2_CID_BASE;
    queryctrl.id < V4L2_CID_LASTP1;
    queryctrl.id++) {
        if (ioctl(fd, VIDIOC_QUERYCTRL, &queryctrl) == 0) {
            if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
                printf("Control %s disabled\n", queryctrl.name);
                continue;
            }
            
            printf("Control %s\n", queryctrl.name);
            
            if (queryctrl.type == V4L2_CTRL_TYPE_MENU)
                enumerateMenu(queryctrl, querymenu);
        } 
        else {
            if (errno == EINVAL) {
                //perror("VIDIOC_QUERYCTRL");
                continue;
            }
            
            perror("VIDIOC_QUERYCTRL");
            exit(EXIT_FAILURE);
        }
    }
    
    for (queryctrl.id = V4L2_CID_PRIVATE_BASE;;
    queryctrl.id++) {
        if (0 == ioctl(fd, VIDIOC_QUERYCTRL, &queryctrl)) {
            if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED)
                continue;
            
            printf("Control %s\n", queryctrl.name);
            
            if (queryctrl.type == V4L2_CTRL_TYPE_MENU)
                enumerateMenu(queryctrl, querymenu);
        } else {
            if (errno == EINVAL)
                break;
            
            perror("VIDIOC_QUERYCTRL");
            exit(EXIT_FAILURE);
        }
    }
}

